"""Evaluate the probability that multiple transfers were bundled in a transaction."""

from forta_agent.transaction_event import TransactionEvent
from web3 import Web3

import ioseeth.indicators.batch
import ioseeth.metrics.probabilities
import ioseeth.options

# CONFIDENCE ##################################################################

def confidence_score(
    log: TransactionEvent,
    w3: Web3,
    min_transfer_count: int=ioseeth.options.MIN_TRANSFER_COUNT,
    min_transfer_total_erc20: int=ioseeth.options.MIN_TRANSFER_TOTAL_ERC20,
    min_transfer_total_native: int=ioseeth.options.MIN_TRANSFER_TOTAL_NATIVE,
    max_batching_fee: int=ioseeth.options.MAX_BATCHING_FEE[1]
) -> float:
    """Evaluate the probability that multiple transfers were bundled in a transaction."""
    _scores = []
    # parse the log
    _block = int(log.block.number)
    _from = str(getattr(log.transaction, 'from_', '')).lower()
    _data = str(getattr(log.transaction, 'data', '')).lower()
    _value = int(getattr(log.transaction, 'value', ''))
    _logs = tuple(log.logs)
    # method selector
    _scores.append(ioseeth.metrics.probabilities.indicator_to_probability(
        indicator=ioseeth.indicators.batch.input_data_has_batching_selector(_data),
        true_score=0.9, # almost certainty
        false_score=0.5)) # not all selectors are in the wordlist: neutral
    # list of recipients and amounts with same length
    _scores.append(ioseeth.metrics.probabilities.indicator_to_probability(
        indicator=ioseeth.indicators.batch.input_data_has_matching_arrays_of_values_and_addresses(data=_data, min_length=min_transfer_count),
        true_score=0.8, # having both lists is necessary, and rarely seen in other transactions
        false_score=0.1)) # without lists of recipients and amounts, there is little chance the contract performs batching
    return ioseeth.metrics.probabilities.conflation(_scores)

# MALICIOUS ###################################################################

#TODO: "to" address keeps tokens (instead of redistributing them)

def malicious_score(
    log: TransactionEvent,
    w3: Web3,
    min_transfer_count: int=ioseeth.options.MIN_TRANSFER_COUNT,
    max_batching_fee: int=ioseeth.options.MAX_BATCHING_FEE[1]
) -> float:
    """Evaluate the provabability that a batch transaction is malicious."""
    _scores = []
    _logs = tuple(log.logs)
    # transfer of amount 0
    _scores.append(ioseeth.metrics.probabilities.indicator_to_probability(
        indicator=ioseeth.indicators.batch.log_has_erc20_transfer_of_null_amount(logs=_logs),
        true_score=0.9, # certainty
        false_score=0.5)) # neutral
    return ioseeth.metrics.probabilities.conflation(_scores)



"""Filter the logs for relevant ERC20 / ERC721 events."""

import copy
import functools
import itertools
import json
import logging

from eth_abi.abi import ABICodec
from eth_utils.abi  import event_abi_to_log_topic
from hexbytes import HexBytes
from web3._utils.abi import build_strict_registry
from web3._utils.events import get_event_data
from web3.exceptions import LogTopicError, MismatchedABI
from web3.types import ABIEvent, LogReceipt

# ABIs ########################################################################

# TODO variants with / without indexing

ERC20_APPROVAL_EVENT = ABIEvent(json.loads('{"name":"Approval","type":"event","anonymous":false,"inputs":[{"indexed":true,"name":"_owner","type":"address"},{"indexed":true,"name":"_spender","type":"address"},{"indexed":false,"name":"_value","type":"uint256"}]}'))
ERC20_TRANSFER_EVENT = ABIEvent(json.loads('{"name":"Transfer","type":"event","anonymous":false,"inputs":[{"indexed":true,"name":"_from","type":"address"},{"indexed":true,"name":"_to","type":"address"},{"indexed":false,"name":"_value","type":"uint256"}]}'))
ERC721_APPROVAL_EVENT = ABIEvent(json.loads('{"name":"Approval","type":"event","anonymous":false,"inputs":[{"indexed":true,"name":"_owner","type":"address"},{"indexed":true,"name":"_approved","type":"address"},{"indexed":true,"name":"_tokenId","type":"uint256"}]}'))
ERC721_APPROVAL_FOR_ALL_EVENT = ABIEvent(json.loads('{"name":"ApprovalForAll","type":"event","anonymous":false,"inputs":[{"indexed":true,"name":"_owner","type":"address"},{"indexed":true,"name":"_operator","type":"address"},{"indexed":false,"name":"_approved","type":"bool"}]}'))
ERC721_TRANSFER_EVENT = ABIEvent(json.loads('{"name":"Transfer","type":"event","anonymous":false,"inputs":[{"indexed":true,"name":"_from","type":"address"},{"indexed":true,"name":"_to","type":"address"},{"indexed":true,"name":"_tokenId","type":"uint256"}]}'))

def _get_input_names(abi: ABIEvent) -> tuple:
    """Extract the name of each input of an event, from its ABI."""
    return tuple(_a.get('name', '') for _a in abi.get('inputs', []))

@functools.lru_cache(maxsize=128)
def _abi_codec() -> ABICodec:
    """Wrapper around the registry for encoding & decoding ABIs."""
    return ABICodec(build_strict_registry())

def _apply_indexation_mask(abi: ABIEvent, mask: tuple) -> ABIEvent:
    """Change the "indexed" field of the ABI according to the mask."""
    _abi = copy.deepcopy(abi)
    for _i in range(len(mask)):
        _abi['inputs'][_i]['indexed'] = mask[_i]
    return _abi

def _generate_all_abi_indexation_variants(abi: ABIEvent) -> dict:
    """Generate all the variants of the ABI by switching each "indexed" field true / false for the inputs."""
    _count = len(abi.get('inputs', ()))
    _indexed = tuple(itertools.product(*(_count * ((True, False), ))))
    _abis = {_c: [] for _c in range(_count + 1)} # order by number of indexed inputs
    for _i in _indexed: # each indexation variant
        _abis[sum(_i)].append(_apply_indexation_mask(abi=abi, mask=_i))
    return _abis

def _generate_the_most_probable_abi_indexation_variants(abi: ABIEvent) -> dict:
    """Generate the most probable variant of the ABI for each count of indexed inputs."""
    _count = len(abi.get('inputs', ()))
    _indexed = tuple((_i * [True] + (_count - _i) * [False]) for _i in range(_count + 1)) # index from left to right, without gaps
    return {sum(_i): _apply_indexation_mask(abi=abi, mask=_i) for _i in _indexed} # order by number of indexed inputs

def _compare_abi_to_log(abi: ABIEvent, log: LogReceipt) -> bool:
    """Returns True if abit and log match, False otherwise."""
    return (
        bool(log['topics'])
        and event_abi_to_log_topic(abi) == log['topics'][0])

# FORMAT ######################################################################

@functools.lru_cache(maxsize=128)
def _get_arg_value(event: 'AttributeDict', name: str) -> str:
    """Extract the value of an event input from its log.""" 
    return str(event.get('args', {}).get(name, ''))

@functools.lru_cache(maxsize=128)
def _get_token_address(event: 'AttributeDict') -> str:
    """Extract the address of the token that emitted the event."""
    return str((event.get('address', '')))

@functools.lru_cache(maxsize=128)
def _parse_event(event: 'AttributeDict', names: tuple) -> dict:
    """Extract the relevant data from a log and format it."""
    return {
        'token': _get_token_address(event=event),
        'from': _get_arg_value(event=event, name=names[0]),
        'to': _get_arg_value(event=event, name=names[1]),
        'value': _get_arg_value(event=event, name=names[2])}

# DECODE ######################################################################

def get_event_data_factory(abi: ABIEvent, codec: ABICodec) -> list:
    """Adapt the parsing logic to a given event."""
    _abi_variants = _generate_the_most_probable_abi_indexation_variants(abi=abi)

    @functools.lru_cache(maxsize=128)
    def _get_event_data(logs: tuple) -> list:
        """Extract event data from the hex log topics."""
        _results = []
        for _log in logs:
            _log.topics = [HexBytes(_topic) for _topic in _log.topics]
            if _compare_abi_to_log(abi=abi, log=_log): # avoid MismatchedABI exception
                _abi = _abi_variants.get(len(_log['topics']) - 1, None) # avoid LogTopicError exception
                if _abi:
                    _results.append(get_event_data(codec, _abi, _log))
        return _results

    return _get_event_data

def parse_logs_factory(abi: ABIEvent=ERC20_TRANSFER_EVENT, codec: ABICodec=_abi_codec()) -> callable:
    """Adapt the parsing logic to a given event."""
    _inputs = _get_input_names(abi)
    _get_event_data = get_event_data_factory(abi=abi, codec=codec)

    @functools.lru_cache(maxsize=128)
    def _parse_logs(logs: tuple) -> tuple:
        """Extract all the event matching a given ABI."""
        _events = _get_event_data(logs=logs)
        # return the args of each event in a dict
        return tuple(_parse_event(event=_e, names=_inputs) for _e in _events)

    return _parse_logs

# SHORTHANDS ##################################################################

filter_logs_for_erc20_transfer_events = parse_logs_factory(abi=ERC20_TRANSFER_EVENT, codec=_abi_codec())

filter_logs_for_erc721_transfer_events = parse_logs_factory(abi=ERC721_TRANSFER_EVENT, codec=_abi_codec())



"""Metadata on the blockchains."""

import json
import os.path

import ioseeth.__utils as __utils

# DATA ########################################################################

with open(os.path.join(__utils.get_data_dir_path(), 'chains.json'), 'r') as __f:
    __chains = json.load(__f)
    NATIVE_TOKENS = {int(__v): __k for __v, __k in __chains.items()}



"""Evaluate the probability that a transaction resulted in an airdrop."""

from forta_agent.transaction_event import TransactionEvent
from web3 import Web3

import ioseeth.indicators.batch
import ioseeth.metrics.probabilities
import ioseeth.options

# CONFIDENCE ##################################################################

def confidence_score(
    log: TransactionEvent,
    w3: Web3,
    min_transfer_count: int=ioseeth.options.MIN_TRANSFER_COUNT,
    min_transfer_total: int=ioseeth.options.MIN_TRANSFER_TOTAL_ERC20
) -> float:
    """Evaluate the probability that a transaction is an airdrop."""
    _scores = []
    _data = str(getattr(log.transaction, 'data', '')).lower()
    _logs = tuple(log.logs)
    # performs token transfers
    _has_token_mint_events = (
        ioseeth.indicators.batch.log_has_multiple_erc20_mint_events(logs=_logs, min_count=min_transfer_count, min_total=min_transfer_total)
        or ioseeth.indicators.batch.log_has_multiple_erc721_mint_events(logs=_logs, min_count=min_transfer_count))
    _scores.append(ioseeth.metrics.probabilities.indicator_to_probability(
        indicator=_has_token_mint_events,
        true_score=0.9, # the tokens were minted
        false_score=0.2)) # could be another standard
    # doesn't have input
    _scores.append(ioseeth.metrics.probabilities.indicator_to_probability(
        indicator=not ioseeth.indicators.batch.input_data_has_array_of_addresses(data=_data, min_length=min_transfer_count),
        true_score=0.6, # not enough to conclude
        false_score=0.4)) # some airdrop functions take inputs
    return ioseeth.metrics.probabilities.conflation(_scores)

# MALICIOUS ###################################################################

# TODO: contract accumulates wealth
# TODO: new contract / new token
# TODO: contract pretends to be a known token (ex: Tether USDT)

def malicious_score(log: TransactionEvent, w3: Web3) -> float:
    """Evaluate the provabability that an airdrop is malicious."""
    _scores = []
    return ioseeth.metrics.probabilities.conflation(_scores)



"""Evaluate the probability that multiple transfers were bundled in a transaction."""

from forta_agent.transaction_event import TransactionEvent
from web3 import Web3

import ioseeth.indicators.generic
import ioseeth.metrics.probabilities
import ioseeth.options

# CONSTANTS ###################################################################

UNUSUAL_OPCODES = ('BLOCKHASH', 'DIFFICULTY')

RED_PILL_OPCODES = ('COINBASE',)

# RED PILL ####################################################################

def is_red_pill(
    bytecode: str,
) -> float:
    """Evaluate the probability that a contract is trying to evade simulation environments."""
    __scores = []
    # opcodes used to detect simulation environments
    __scores.append(ioseeth.metrics.probabilities.indicator_to_probability(
        indicator=ioseeth.indicators.generic.bytecode_has_specific_opcodes(bytecode=bytecode, opcodes=RED_PILL_OPCODES, check=any),
        true_score=0.7,
        false_score=0.5))
    return ioseeth.metrics.probabilities.conflation(__scores)



"""Evaluate the probability that multiple transfers were bundled in a transaction."""

from forta_agent.transaction_event import TransactionEvent
from web3 import Web3

import ioseeth.indicators.proxy
import ioseeth.metrics.probabilities
import ioseeth.options

# PROXY #######################################################################

#TODO improve bytecode disassembly: wrong opcode
#TODO delegatecall opcode appears after disassembly when it's not used in original sources...

def is_redirecting_execution_to_another_contract(
    data: str,
    bytecode: str
) -> float:
    """Evaluate the probability that a given contract redirects the execution to another contract."""
    __scores = []
    # uses staticcall/delegatecall/callcode
    # __scores.append(ioseeth.metrics.probabilities.indicator_to_probability(
    #     indicator=ioseeth.indicators.proxy.bytecode_redirects_execution(bytecode=bytecode),
    #     true_score=1, # has instructions that definitely redirect to another contract
    #     false_score=0.3)) # can still use the logic of another contract without modifying its state 
    # list of recipients and amounts with same length
    __scores.append(ioseeth.metrics.probabilities.indicator_to_probability(
        indicator=not ioseeth.indicators.generic.bytecode_has_implementation_for_transaction_selector(bytecode=bytecode, data=data),
        true_score=0.7, # execution goes through the fallback, possibly to another contract
        false_score=0.1))
    return ioseeth.metrics.probabilities.conflation(__scores)

def is_standard_proxy(
    data: str,
    bytecode: str
) -> float:
    """Evaluate the probability that the given contract is a proxy."""
    __scores = []
    # uses staticcall/delegatecall/callcode
    __scores.append(ioseeth.metrics.probabilities.indicator_to_probability(
        indicator=is_redirecting_execution_to_another_contract(data=data, bytecode=bytecode),
        true_score=0.6, # acts like a proxy, but may still be another type of contract
        false_score=0.)) # proxies can't work without redirecting execution
    # list of recipients and amounts with same length
    __scores.append(ioseeth.metrics.probabilities.indicator_to_probability(
        indicator=ioseeth.indicators.proxy.bytecode_uses_standard_proxy_slots(bytecode=bytecode),
        true_score=0.8, # very little chance another type of con
        false_score=0.3))
    return ioseeth.metrics.probabilities.conflation(__scores)

# ISSUES ######################################################################

def has_broken_proxy_implementation(
    bytecode: str
) -> float:
    """Evaluate the probability the proxy is not properly written."""
    __scores = []
    # has logic slots from several standards
    __scores.append(ioseeth.metrics.probabilities.indicator_to_probability(
        indicator=ioseeth.indicators.proxy.bytecode_has_proxy_slots_from_several_standards(bytecode=bytecode),
        true_score=0.6,
        false_score=0.5))
    return ioseeth.metrics.probabilities.conflation(__scores)



"""Evaluate the probability that a transaction resulted in transfers of NFTs."""

from forta_agent.transaction_event import TransactionEvent
from web3 import Web3

import ioseeth.indicators.batch
import ioseeth.metrics.probabilities
import ioseeth.options

# CONFIDENCE ##################################################################

def confidence_score(
    log: TransactionEvent,
    w3: Web3,
    min_transfer_count: int=ioseeth.options.MIN_TRANSFER_COUNT,
) -> float:
    """Evaluate the probability that a transaction handled NFT tokens."""
    _scores = []
    _logs = tuple(log.logs)
    # events
    _scores.append(ioseeth.metrics.probabilities.indicator_to_probability(
        indicator=ioseeth.indicators.batch.log_has_multiple_erc721_transfer_events(logs=_logs, min_count=min_transfer_count),
        true_score=0.9, # certainty
        false_score=0.2)) # the token could follow another std
    return ioseeth.metrics.probabilities.conflation(_scores)

# MALICIOUS ###################################################################

def malicious_score(log: TransactionEvent, w3: Web3) -> float:
    """Evaluate the provabability that a NFT transaction is malicious."""
    _scores = []
    return ioseeth.metrics.probabilities.conflation(_scores)



"""Indicators on token contracts."""

import functools

import ioseeth.indicators.generic
import ioseeth.parsing.abi
import ioseeth.parsing.bytecode
import ioseeth.parsing.inputs

# CONSTANTS ###################################################################

ERC777_ABI = ioseeth.parsing.abi.load(path='interfaces/IERC777.json')
ERC20_ABI = ioseeth.parsing.abi.load(path='token/ERC20/ERC20.json')
ERC721_ABI = ioseeth.parsing.abi.load(path='token/ERC721/ERC721.json')
ERC1155_ABI = ioseeth.parsing.abi.load(path='token/ERC1155/ERC1155.json')

# ERC-20 ######################################################################

@functools.lru_cache(maxsize=128)
def bytecode_has_erc20_interface(bytecode: str, abi: dict=ERC20_ABI, threshold: float=0.8) -> bool:
    return ioseeth.indicators.generic.bytecode_has_specific_interface(bytecode=bytecode, abi=abi, threshold=threshold)

# ERC-721 #####################################################################

@functools.lru_cache(maxsize=128)
def bytecode_has_erc721_interface(bytecode: str, abi: dict=ERC721_ABI, threshold: float=0.8) -> bool:
    return ioseeth.indicators.generic.bytecode_has_specific_interface(bytecode=bytecode, abi=abi, threshold=threshold)

# ERC-777 #####################################################################

@functools.lru_cache(maxsize=128)
def bytecode_has_erc777_interface(bytecode: str, abi: dict=ERC777_ABI, threshold: float=0.8) -> bool:
    return ioseeth.indicators.generic.bytecode_has_specific_interface(bytecode=bytecode, abi=abi, threshold=threshold)

# ERC-1155 ####################################################################

@functools.lru_cache(maxsize=128)
def bytecode_has_erc1155_interface(bytecode: str, abi: dict=ERC1155_ABI, threshold: float=0.8) -> bool:
    return ioseeth.indicators.generic.bytecode_has_specific_interface(bytecode=bytecode, abi=abi, threshold=threshold)

# ANY TOKEN ###################################################################

@functools.lru_cache(maxsize=128)
def bytecode_has_any_token_interface(bytecode: str, threshold: float=0.8) -> bool:
    return (
        bytecode_has_erc20_interface(bytecode=bytecode, threshold=threshold)
        or bytecode_has_erc721_interface(bytecode=bytecode, threshold=threshold)
        or bytecode_has_erc777_interface(bytecode=bytecode, threshold=threshold)
        or bytecode_has_erc1155_interface(bytecode=bytecode, threshold=threshold))



"""Generic indicators for smart contracts."""

import functools

import ioseeth.parsing.abi
import ioseeth.parsing.bytecode
import ioseeth.parsing.inputs

# INTERFACES ##################################################################

@functools.lru_cache(maxsize=128)
def bytecode_has_specific_interface(bytecode: str, abi: dict, threshold: float=0.8, raw: bool=False) -> bool:
    __selectors = ioseeth.parsing.bytecode.get_function_selectors(bytecode=bytecode, raw=raw)
    __interface = tuple(abi.calculate_all_selectors(abi=abi, target='function').values())
    return (sum(__s in __interface for __s in __selectors) / len(__interface)) >= threshold # only requires to have threshold % of the interface

# CONFLICTS ###################################################################

@functools.lru_cache(maxsize=128)
def bytecode_has_implementation_for_transaction_selector(bytecode: str, data: str) -> bool:
    return (
        not data
        or (len(data) < 6)
        or (ioseeth.parsing.inputs.get_function_selector(data=data) in ioseeth.parsing.bytecode.get_function_selectors(bytecode=bytecode)))

# UNUSUAL CODING PATTERNS #####################################################

@functools.lru_cache(maxsize=128)
def bytecode_has_specific_opcodes(bytecode: str, opcodes: str, check: callable=any) -> bool:
    return check(_o in bytecode for _o in opcodes)



"""Evaluate the probability that multiple transfers were bundled in a transaction."""

from forta_agent.transaction_event import TransactionEvent
from web3 import Web3

import ioseeth.indicators.proxy
import ioseeth.indicators.token
import ioseeth.metrics.probabilities
import ioseeth.metrics.normal.proxy
import ioseeth.options

# HIDDEN PROXY ################################################################

# is redirecting execution ?
# no => out
# yes
    # is standard proxy ?
        # yes
            # is their an address in the implementation slot?
                # yes
                    # is the implementation contract empty?
                        # yes => weird (check contract age)
                        # no => is it the one called?
                # no => weird (view contract age)
        # no
            # is it a token?
                # yes => weird (token + delegate)

def is_hidden_proxy(
    data: str,
    bytecode: str
) -> float:
    """Evaluate that a contract is redirecting execution."""
    __scores = []
    # requirement: must redirect execution
    __scores.append(ioseeth.metrics.probabilities.indicator_to_probability(
        indicator=ioseeth.metrics.normal.proxy.is_redirecting_execution_to_another_contract(data=data, bytecode=bytecode) >= 0.7,
        true_score=0.5, # does not say whether the contract is malicious
        false_score=0.)) # cannot be a hidden a proxy if it's not even a proxy
    # std proxies... should follow standards
    # tokens shouldn't redirect
    __scores.append(ioseeth.metrics.probabilities.indicator_to_probability(
        indicator=ioseeth.indicators.token.bytecode_has_any_token_interface(bytecode=bytecode, threshold=0.9),
        true_score=0.8, # tokens should never happen redirect
        false_score=0.5)) # there are other types of hidden proxies
    return ioseeth.metrics.probabilities.conflation(__scores)



"""Handle ABIs."""

import json
import os.path

import eth_utils

import ioseeth.__utils as __utils

# SIGNATURE ###################################################################

def format_signature(name: str, types: tuple) -> str:
    """Compose a function / event / error signature from its name and argument types."""
    return '{name}({arguments})'.format(
        name=name,
        arguments=','.join(types))

# HASH ########################################################################

def calculate_selector(signature: str) -> str:
    """Compute the method selector for a single signature."""
    return (eth_utils.crypto.keccak(text=signature).hex().lower())[:8] # "0x" prefix + 4 bytes

# ABI #########################################################################

def load(path: str) -> dict:
    """Load an ABI from the references on disk."""
    with open(os.path.join(__utils.get_data_dir_path(), path), 'r') as __f:
        return json.load(__f)

def calculate_all_selectors(abi: list, target: str='function') -> dict:
    """Compute the selector of each function and returns a dictionary {signature => selector}."""
    __result = {}
    for __o in abi:
        if __o.get('type', '') == target:
            __signature = format_signature(
                name=__o.get('name', ''),
                types=(__a.get('type', '') for __a in __o.get('inputs', [])))
            __result[__signature] = selector(signature=__signature)
    return __result



"""Base indicators on transactions and their metadata."""

import functools
import itertools

from web3 import Web3

import ioseeth.indicators.wordlists as wordlists
import ioseeth.parsing.abi as abi
import ioseeth.parsing.balances as balances
import ioseeth.parsing.events as events
import ioseeth.parsing.inputs as inputs

# SELECTORS INDICATORS ########################################################

KNOWN_SIGNATURES = (
    wordlists.generate_signature_wordlist(pattern=wordlists.PATTERNS[0], verbs=wordlists.VERBS, adjectives=wordlists.ADJECTIVES, tokens=wordlists.TOKENS, nouns=wordlists.NOUNS, args=wordlists.ARGS)
    + wordlists.generate_signature_wordlist(pattern=wordlists.PATTERNS[1], verbs=wordlists.VERBS, adjectives=wordlists.ADJECTIVES, tokens=wordlists.TOKENS, nouns=wordlists.NOUNS, args=wordlists.ARGS))

KNOWN_SELECTORS = {abi.calculate_selector(_s): _s for _s in KNOWN_SIGNATURES}

@functools.lru_cache(maxsize=128)
def input_data_has_batching_selector(data: str, known: tuple=tuple(KNOWN_SELECTORS)) -> bool:
    return data[:10].lower() in known # selector => signature mapping

# INPUTS INDICATORS ###########################################################

@functools.lru_cache(maxsize=128)
def input_data_has_array_of_addresses(data: str, min_length: int) -> bool:
    return len(inputs.get_array_of_address_candidates(data=data, min_length=min_length)) > 0 # at least one candidate array

@functools.lru_cache(maxsize=128)
def input_data_has_array_of_values(data: str, min_length: int) -> bool:
    return len(inputs.get_array_of_value_candidates(data=data, min_length=min_length)) > 0 # at least one candidate array

@functools.lru_cache(maxsize=128)
def input_data_has_matching_arrays_of_values_and_addresses(data: str, min_length: int) -> bool:
    return bool(inputs.get_matching_arrays_of_address_and_value(data=data, min_length=min_length))

# EVENTS INDICATORS ###########################################################

# TODO: add ERC1155

@functools.lru_cache(maxsize=128)
def log_has_multiple_erc20_transfer_events(logs: tuple, min_count: int, min_total: int) -> bool:
    _events = events.filter_logs_for_erc20_transfer_events(logs=logs)
    _amounts = [int(_e['value']) for _e in _events]
    return len(_events) >= min_count and sum(_amounts) >= min_total

@functools.lru_cache(maxsize=128)
def log_has_multiple_erc20_mint_events(logs: tuple, min_count: int, min_total: int) -> bool:
    _events = events.filter_logs_for_erc20_transfer_events(logs=logs)
    _amounts = [int(_e['value']) for _e in _events]
    _origins = [int(_e['from'], 16) == 0 for _e in _events] # creation / minting of tokens
    return len(_events) >= min_count and sum(_amounts) >= min_total and all(_origins)

@functools.lru_cache(maxsize=128)
def log_has_erc20_transfer_of_null_amount(logs: tuple) -> bool:
    _events = events.filter_logs_for_erc20_transfer_events(logs=logs)
    _amounts = [int(_e['value']) for _e in _events]
    return any([_a == 0 for _a in _amounts])

@functools.lru_cache(maxsize=128)
def log_has_multiple_erc721_transfer_events(logs: tuple, min_count: int) -> bool:
    return len(events.filter_logs_for_erc721_transfer_events(logs=logs)) >= min_count

@functools.lru_cache(maxsize=128)
def log_has_multiple_erc721_mint_events(logs: tuple, min_count: int) -> bool:
    _events = events.filter_logs_for_erc721_transfer_events(logs=logs)
    _origins = [int(_e['from'], 16) == 0 for _e in _events] # creation / minting of tokens
    return len(_events) >= min_count and all(_origins)

# VALUE INDICATORS ###########################################################

@functools.lru_cache(maxsize=128)
def transaction_value_matches_input_arrays(value: int, data: str, min_count: int, tolerance: int) -> bool:
    _args = inputs.get_matching_arrays_of_address_and_value(data=data, min_length=min_count)
    _sums = [sum(_a[1]) for _a in _args] # sum each candidate array of amounts
    _tests = [abs(value - _s) <= tolerance for _s in _sums] # account for a service fee with the tolerance
    return any(_tests) # one of the input arrays has a sum equal to the transaction value

# BALANCES INDICATORS #########################################################

@functools.lru_cache(maxsize=128)
def multiple_native_token_balances_changed(w3: Web3, data: str, block: int, min_count: int, min_total: int) -> bool:
    _recipients = itertools.chain.from_iterable(inputs.get_array_of_address_candidates(data=data, min_length=min_count)) # list of candidates = list of lists
    _deltas = [balances.get_balance_delta(w3=w3, address=_a, block=block) for _a in _recipients] # _recipients is now a flat list
    return len(_deltas) >= min_count and sum(_deltas) >= min_total

@functools.lru_cache(maxsize=128)
def native_token_balance_changed(w3: Web3, address: str, block: int, tolerance: int=10**17) -> bool:
    return balances.get_balance_delta(w3=w3, address=address, block=block) > tolerance # in case the contract has a fee, set to 0.1 EHT by default



"""Evaluate the probability that a transaction resulted in transfers of ERC20 tokens."""

from forta_agent.transaction_event import TransactionEvent
from web3 import Web3

import ioseeth.indicators.batch
import ioseeth.metrics.probabilities
import ioseeth.options

# CONFIDENCE ##################################################################

# TODO: add ERC115

def confidence_score(
    log: TransactionEvent,
    w3: Web3,
    min_transfer_count: int=ioseeth.options.MIN_TRANSFER_COUNT,
    min_transfer_total: int=ioseeth.options.MIN_TRANSFER_TOTAL_ERC20
) -> float:
    """Evaluate the probability that a transaction handled ERC20 tokens."""
    _scores = []
    _logs = tuple(log.logs)
    # events
    _scores.append(ioseeth.metrics.probabilities.indicator_to_probability(
        indicator=ioseeth.indicators.batch.log_has_multiple_erc20_transfer_events(logs=_logs, min_count=min_transfer_count, min_total=min_transfer_total),
        true_score=0.9, # certainty
        false_score=0.2)) # the token could follow another std
    return ioseeth.metrics.probabilities.conflation(_scores)

# MALICIOUS ###################################################################

# TODO: the ERC20 balance of the contract increased

def malicious_score(log: TransactionEvent, w3: Web3) -> float:
    """Evaluate the provabability that an ERC20 transaction is malicious."""
    _scores = []
    _logs = tuple(log.logs)
    # transfer of amount 0
    _scores.append(ioseeth.metrics.probabilities.indicator_to_probability(
        indicator=ioseeth.indicators.batch.log_has_erc20_transfer_of_null_amount(logs=_logs),
        true_score=0.9, # certainty
        false_score=0.5)) # neutral
    return ioseeth.metrics.probabilities.conflation(_scores)



"""Indicators on proxy contracts."""

import functools
import web3

import ioseeth.parsing.bytecode

# CONSTANTS ###################################################################

DELEGATE_OPCODES = ('DELEGATECALL',)

LOGIC_SLOTS = {
    # bytes32(uint256(keccak256('eip1967.proxy.implementation')) - 1)
    'erc-1967': '360894a13ba1a3210667c828492db98dca3e2076cc3735a920a3ca505d382bbc',
    # keccak256("org.zeppelinos.proxy.implementation")
    'zeppelinos': '7050c9e0f4ca769c69bd3a8ef740bc37934f8e2c036e5a723fd8ee048ed3f8c3',
    # keccak256("PROXIABLE")
    'erc-1822': 'c5f16f0fcc639fa48a6947836d9850f504798523bf8c9a3a87d5876cf622bcf7',}

BEACON_SLOTS = {
    # bytes32(uint256(keccak256('eip1967.proxy.beacon')) - 1)
    'erc-1967': 'a3f0ad74e5423aebfd80d3ef4346578335a9a72aeaee59ff6cb3582b35133d50',}

INTERFACES = {
    'erc-1167': (
        # bytes4(keccak256("implementation()"))
        '5c60da1b',
        # bytes4(keccak256("childImplementation()"))
        'da525716',),
    'erc-1167': (
        # bytes4(keccak256("implementation()"))
        '5c60da1b',),
    'gnosis-safe': (
        # Gnosis Safe: bytes4(keccak256("masterCopy()"))
        'a619486e',),
    'comptroller': (
        # Comptroller: bytes4(keccak256("comptrollerImplementation()"))
        'bb82aa5e',),}

# PROXY #######################################################################

#TODO improve bytecode disassembly: wrong opcode
#TODO delegatecall opcode appears after disassembly when it's not used in original sources...

@functools.lru_cache(maxsize=128)
def bytecode_redirects_execution(bytecode: str, opcodes: tuple=DELEGATE_OPCODES) -> bool:
    return any(_o in bytecode for _o in opcodes)

# STANDARDS ###################################################################

@functools.lru_cache(maxsize=128)
def bytecode_uses_standard_proxy_slots(bytecode: str, standards: dict=LOGIC_SLOTS) -> bool:
    return any(_slot in bytecode for _slot in standards.values())

@functools.lru_cache(maxsize=128)
def bytecode_has_proxy_slots_from_several_standards(bytecode: str, standards: dict=LOGIC_SLOTS) -> bool:
    return sum(_slot in bytecode for _slot in standards.values()) > 1

# LOGIC CONTRACT ##############################################################

@functools.lru_cache(maxsize=128)
def storage_logic_addresses(w3: web3.Web3, address: str, bytecode: str, standards: dict=LOGIC_SLOTS) -> bool:
    _slots = ioseeth.parsing.bytecode.get_storage_slots(bytecode=bytecode)
    _values = (w3.eth.get_storage_at(address, _s) for _s in standards.values() if _s in _slots)
    return ('0x' + (_v.hex())[26:] for _v in _values if int(_v.hex(), 16) > 0)



"""Track the evolution of native token balances."""

import functools

from web3 import Web3

# DELTA #######################################################################

@functools.lru_cache(maxsize=128)
def get_balance_delta(provider: Web3, address: str, block: int) -> int:
    """Calculate the difference in balance before / after a given block."""
    _before = _after = 0
    if address:
        _before = provider.eth.get_balance(Web3.toChecksumAddress(address), block - 1)
        _after = provider.eth.get_balance(Web3.toChecksumAddress(address), block)
    return _after - _before

@functools.lru_cache(maxsize=128)
def get_balance_deltas(provider: Web3, addresses: list, block: int) -> dict:
    """List all the addresses that sustained a balance change."""
    _deltas = {_a: get_balance_delta(provider=provider, address=_a, block=block) for _a in addresses}
    return {_a: _d for _a, _d in _deltas.items() if abs(_d) > 0}



"""Parse EVM bytecode and extract the key elements.

Follows the decomposition detailed in the article by OpenZeppelin:
https://blog.openzeppelin.com/deconstructing-a-solidity-contract-part-i-introduction-832efd2d7737
"""

import functools
import re

# GENERIC #####################################################################

@functools.lru_cache(maxsize=128)
def is_raw_hex(bytecode: str) -> bool:
    """Check whether the bytecode is a raw hexadecimal string or a sequence of opcodes."""
    return not (
        isinstance(bytecode, list)
        or isinstance(bytecoden, tuple)
        or (isinstance(bytecode, str) and ' ' in bytecode))

# REGEX #######################################################################

@functools.lru_cache(maxsize=128)
def selector_regex(raw: bool=False) -> str:
    """Regex matching a function selector, in the bytecode."""
    return (
        r'63([0-9a-f]{8})14' if raw
        else r'PUSH4\s+([0-9a-f]{8})')

@functools.lru_cache(maxsize=128)
def address_regex(raw: bool=False) -> str:
    """Regex matching an address, in the bytecode."""
    return (
        r'73([0-9a-f]{40})' if raw
        else r'PUSH20\s+([0-9a-f]{40})')

@functools.lru_cache(maxsize=128)
def storage_slot_regex(raw: bool=False) -> str:
    """Regex matching a storage slot, in the bytecode."""
    return (
        r'7f([0-9a-f]{64})' if raw
        else r'PUSH32\s+([0-9a-f]{64})')

# SELECTORS ###################################################################

@functools.lru_cache(maxsize=128)
def get_function_selectors(bytecode: str, raw: bool=False) -> tuple:
    """Get all the function selectors from the hub portion of the bytecode."""
    _r = re.compile(selector_regex(raw=raw), re.IGNORECASE)
    return tuple(set(_r.findall(bytecode)))

# STORAGE #####################################################################

#TODO SLOAD could use a computed address instead of a hardcoded one => fetching only the 32 bytes words is too naive

@functools.lru_cache(maxsize=128)
def get_storage_slots(bytecode: str, raw: bool=False) -> str:
    """Get all the storage slots used in the bytecode."""
    _r = re.compile(storage_slot_regex(raw=raw), re.IGNORECASE)
    return tuple(set(_r.findall(bytecode)))



"""Parse HEX input data.

In particular, look for arrays of addresses and amounts.

Address and value regex are designed to be exclusive.
This avoids matching the address when searching for values and vice-versa.

All array functions are generic:
The result depends on the functions given as argument to process each element.
"""

import functools
import re

# GENERIC #####################################################################

def chunk(l, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(l), n):
        yield l[i:i + n]

# DECODING ####################################################################

@functools.lru_cache(maxsize=128)
def max_array_length(data: str) -> int:
    """Calculate the max length of an array based on the length of the data string."""
    # 74 = 10 + 64 = prefix + selector + 32 bytes (uint256) for the array length
    return int((len(data) - 74) / 64) # negative or null if the data string is not long enough

@functools.lru_cache(maxsize=128)
def is_valid_address(data: str) -> bool:
    """Checks whether a hex string is mostly non-zero."""
    return len(hex(int(data, 16))) > 32 # counting the prefix 0x

@functools.lru_cache(maxsize=128)
def is_valid_value(data: str) -> bool:
    """Checks whether a hex string represents a meaningful monetary amount."""
    return len(hex(int(data, 16))) <= 32 # counting the prefix 0x

@functools.lru_cache(maxsize=128)
def is_valid_array(data: str, check: callable, length: int) -> bool:
    """Checks whether a hex string is an array."""
    _chunks = list(chunk(data, 64))
    _valid = len(data) >= 128 # array length + at least 1 element = 64 + 64 = 2 * 32 bytes
    _valid = _valid and len(data) % 64 == 0 # correct ABI encoding
    _valid = _valid and int(_chunks[0], 16) == len(_chunks) - 1 # the first word is the array length
    _valid = _valid and int(_chunks[0], 16) >= length # ignore small arrays
    _valid = _valid and all([check(_c) for _c in _chunks[1:]]) # all the elements are addresses
    return _valid

@functools.lru_cache(maxsize=128)
def parse_address(data: str) -> str:
    """Format a raw word into a 20 byte address."""
    return '0x' + data[24:]

@functools.lru_cache(maxsize=128)
def parse_value(data: str) -> int:
    """Format a raw word into an integer."""
    return int(data, 16)

@functools.lru_cache(maxsize=128)
def parse_array(data: str, parse_element: callable) -> list:
    """Format raw input data into a list of 20 bytes addresses."""
    _chunks = list(chunk(data, 64))
    return [parse_element(_c) for _c in _chunks[1:]] # omit the array length

# REGEX #######################################################################

@functools.lru_cache(maxsize=128)
def address_regex() -> str:
    """Regex matching a single address."""
    return r'0{24}[0-9a-f]{40}'

@functools.lru_cache(maxsize=128)
def value_regex() -> str:
    """Regex matching a single value."""
    return r'0{34}[0-9a-f]{30}' # 10^12 amount with 24 decimals is 0xc097ce7bc90715b34b9f1000000000, length 30

@functools.lru_cache(maxsize=128)
def array_length_regex(length: int, exact: bool=False) -> str:
    """Regex matching a low number that could be an array length."""
    _length = hex(length)[2:] # remove the prefix 0x
    _regex = '0' * (64 - len(_length)) # 0 padding on the left
    _regex += _length if exact else f'[0-9a-f]{{{len(_length)}}}'
    return _regex

@functools.lru_cache(maxsize=128)
def array_regex(length: int, element_regex: str) -> str:
    """Regex matching an entire array of addresses."""
    _length_re = array_length_regex(length=length, exact=True)
    _elements_re = '(?:' + element_regex + f'){{{length}}}' # do not capture elements individually
    return '(' + _length_re + _elements_re + ')' # only capture the whole array as a group

# PARSING #####################################################################

@functools.lru_cache(maxsize=128)
def get_function_selector(data: str) -> str:
    """Extract the function selector from the input data."""
    return data[2:10]

@functools.lru_cache(maxsize=128)
def get_array_length_candidates(data: str) -> list:
    """Extract the words in the data that could encode an array length."""
    _limit = max_array_length(data)
    _chunks = [int(_c, 16) for _c in chunk(data[10:], 64)] # ignore the prefix and selector
    return list(set([_c for _c in _chunks if _c <= _limit])) # remove repeats

@functools.lru_cache(maxsize=128)
def get_array_candidates(data: str, element_regex: str, element_check: callable, parse_element: callable, min_length: int=4) -> list:
    """Extract the address & value arrays from the hex input."""
    _arrays = []
    _length_candidates = get_array_length_candidates(data)
    for _l in _length_candidates:
        _array_re = re.compile(array_regex(length=_l, element_regex=element_regex))
        _array_candidates = _array_re.findall(data.lower())
        for _a in _array_candidates:
            if is_valid_array(data=_a, check=element_check, length=min_length):
                _arrays.append(parse_array(_a, parse_element))
    return _arrays

# HELPERS #####################################################################

@functools.lru_cache(maxsize=128)
def get_array_of_address_candidates(data: str, min_length: int=4) -> list:
    """Extract the address arrays from the hex input."""
    return get_array_candidates(
        data=data,
        element_regex=address_regex(),
        element_check=is_valid_address,
        parse_element=parse_address,
        min_length=min_length)

@functools.lru_cache(maxsize=128)
def get_array_of_value_candidates(data: str, min_length: int=4) -> list:
    """Extract the value arrays from the hex input."""
    return get_array_candidates(
        data=data,
        element_regex=value_regex(),
        element_check=is_valid_value,
        parse_element=parse_value,
        min_length=min_length)

@functools.lru_cache(maxsize=128)
def get_matching_arrays_of_address_and_value(data: str, min_length: int=4) -> list:
    """Extract arrays of addresses and values when they match."""
    _addresses = get_array_of_address_candidates(data=data, min_length=min_length)
    _values = get_array_of_value_candidates(data=data, min_length=min_length)
    return [(_a, _v) for _a in _addresses for _v in _values if len(_a) == len(_v)]



"""Generate potential selectors for the target batching methods
Ex:
- multisendEther(address[],uint256[])
- bulkTransferToken(address,address[],uint[])
"""

import itertools

# DEFAULT ARGUMENTS ###########################################################

# TODO: add optional suffix "from"
# TODO: add optional prefix "safe"
# TODO: add optional argument "_from"
# TODO: "transfer(address[],address[],address[],uint256[])" => transfer(token,from,to,amount)
PATTERNS = (
    '{verb}{adjective}{token}{noun}{args}',
    '{adjective}{verb}{token}{noun}{args}',)

# TODO: sort signatures by action (batch, airdrop, etc)
VERBS = (
    '',
    'Multisend',
    'Disperse',
    'Dispatch',
    'Send',
    'Batch',
    'Bundle',
    'Multicall')

ADJECTIVES = (
    '',
    'Multi',
    'Multiple',
    'Bulk',
    'Batched',
    'Bundled',
    'Batch',
    'Bundle',
    'Mass')

# TODO: sort signatures by token
TOKENS = ('', 'ETH', 'Eth', 'Ether', 'Token', 'Coin', 'NFT', 'Nft', 'ERC20', 'Erc20', 'ERC721', 'Erc721')

NOUNS = ('', 'Sender', 'Transfer', 'Transaction')

# TODO: add argument pattern
# TODO: single value for all addresses
ARGS = (
    '(address,address[],uint256[])',
    '(address,uint256[],address[])',
    '(address[],uint256[],address)',
    '(uint256[],address[],address)',
    '(address[],uint256[])',
    '(uint256[],address[])',
    '(address,address[],uint[])',
    '(address,uint[],address[])',
    '(address[],uint[],address)',
    '(uint[],address[],address)',
    '(address[],uint[])',
    '(uint[],address[])')

# WORDLIST GENERATION #########################################################

def generate_signature_wordlist(
    pattern: tuple=PATTERNS[0],
    verbs: tuple=VERBS,
    adjectives: tuple=ADJECTIVES,
    tokens: tuple=TOKENS,
    nouns: tuple=NOUNS,
    args: tuple=ARGS
) -> list:
    """Generate a list of plausible method signatures."""
    _signatures = []
    for _a in itertools.product(verbs, adjectives, tokens, nouns, args):
        _signature = pattern.format(verb=_a[0], adjective=_a[1], token=_a[2], noun=_a[3], args=_a[-1])
        _signature = _signature[0].lower() + _signature[1:] # camel case
        _signatures.append(_signature)
    return _signatures



import os.path

# DATA ########################################################################

def get_data_dir_path() -> str:
    """Returns the path to the directory containing the data files."""
    return os.path.join(os.path.abspath(os.path.dirname(__file__)), './data/')



"""Use probability theory to score the transactions.

Given a set of probabilities {p_i}, adding another probability p will:
- leave the conflation score unchanged if p = 0.5
- increase the score if p > 0.5
- decrease the score otherwise
"""

import functools
import operator

# CAST ########################################################################

def indicator_to_probability(indicator: bool, true_score: float, false_score: float) -> float:
    """Cast a boolean indicator into a float probability."""
    return float(indicator) * true_score + (1. - float(indicator)) * false_score

# COMBINE #####################################################################

def conflation(scores: list) -> float:
    """Combine several probability scores into one, using the conflation function."""
    _inverse_scores = [1. - _x for _x in scores] # inverse probabilities
    return (
        functools.reduce(operator.mul, scores, 1.)
        / (functools.reduce(operator.mul, scores, 1.) + functools.reduce(operator.mul, _inverse_scores, 1.)))



"""Evaluate the probability that a transaction resulted in transfers of native tokens."""

from forta_agent.transaction_event import TransactionEvent
from web3 import Web3

import ioseeth.indicators.batch
import ioseeth.metrics.probabilities
import ioseeth.options

# CONFIDENCE ##################################################################

def confidence_score(
    log: TransactionEvent,
    w3: Web3,
    min_transfer_count: int=ioseeth.options.MIN_TRANSFER_COUNT,
    min_transfer_total: int=ioseeth.options.MIN_TRANSFER_TOTAL_NATIVE,
    max_batching_fee: int=ioseeth.options.MAX_BATCHING_FEE[1]
) -> float:
    """Evaluate the probability that a transaction resulted in transfers of native tokens."""
    _scores = []
    _block = int(log.block.number)
    _from = str(getattr(log.transaction, 'from_', '')).lower()
    _data = str(getattr(log.transaction, 'data', '')).lower()
    _value = int(getattr(log.transaction, 'value', ''))
    # "from" contract balance significantly changed
    _scores.append(ioseeth.metrics.probabilities.indicator_to_probability(
        indicator=_value >= max_batching_fee, # mvt below 0.1 ETH are ignored
        true_score=0.5, # required, but not conclusive
        false_score=0.1)) # certainty: no batch transfer without updating the sender's balance
    # check whether the transaction value is sprayed (split and sent to multiple recipients)
    _scores.append(ioseeth.metrics.probabilities.indicator_to_probability(
        indicator=ioseeth.indicators.batch.transaction_value_matches_input_arrays(value=_value, data=_data, min_count=min_transfer_count, tolerance=max_batching_fee),
        true_score=0.8, # very likely: the transaction value is split among recipients specified in the inputs
        false_score=0.2)) # addresses specified outside of the inputs could have had their balance changed
    return ioseeth.metrics.probabilities.conflation(_scores)

# MALICIOUS ###################################################################

def malicious_score(
    log: TransactionEvent,
    w3: Web3,
    min_transfer_count: int=ioseeth.options.MIN_TRANSFER_COUNT,
    max_batching_fee: int=ioseeth.options.MAX_BATCHING_FEE[1]
) -> float:
    """Evaluate the provabability that the transfer is malicious."""
    _scores = []
    return ioseeth.metrics.probabilities.conflation(_scores)
